# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Temporary module to put domain related classes."""


class Service:  # pylint:disable=too-few-public-methods
  """TODO: Domain service declaration."""


class Repository:  # pylint:disable=too-few-public-methods
  """TODO: Domain model entity repository."""


class QuerySet:  # pylint:disable=too-few-public-methods
  """TODO: Domain model query set application."""


class Error(Exception):
  """DDD Errors"""


class DomainError(Error):
  """Entities, repositories and services base error for user cases.
  This exceptions does not break the workflow."""


class NoIdentityError(DomainError):
  """Notify the use"""

  def __init__(self, cls):
    super().__init__(cls)
    self._cls = cls

  def __str__(self):
    return f'{self._cls.__qualname__} without identity'


class RepositoryError(DomainError):
  """Find, Store, Delete, Remove errors"""


class ServiceError(DomainError):
  """Service logic error."""


class NotFoundError(RepositoryError):
  """Raised when Repository.Find doest not foun.
  Connection and transaction errors not included.
  """



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""This module is under construction.
It's used to mark the domain role for classes and models defined in the domain.
"""

from __future__ import annotations

from abc import ABC, ABCMeta, abstractmethod
from functools import lru_cache, wraps
from sys import version_info
from types import CodeType, FunctionType
from typing import (Callable, Generic, NoReturn, Type, TypeVar, _GenericAlias,
                    _SpecialForm, cast, get_type_hints)


class Stuff(ABC):
  def __init__(self, **kwargs):
    self.__dict__.update(kwargs)


T = TypeVar('T')
ID = TypeVar('ID')
ORD = -3

# TODO: remove after move new_ddd support
if version_info.minor > 9:
  ORD -= 1


class MetaEntity(ABCMeta):
  def __init__(cls, cname, bases, namespace):
    if '__annotations__' in namespace:
      initlines = ['def __init__(self']
      initbody = []
      slots = []
      idname = None
      for name, kind in get_type_hints(cls).items():
        if isinstance(kind, _GenericAlias):
          kind = cast(_GenericAlias, kind).__origin__
        initlines.append(f', {name}: {kind.__name__}')
        initbody.append(f'\n  self.{name} = {name}')
        slots.append(name)
        if isinstance(kind, IdentityAlias):
          idname = name
      initlines.append('):\n  """Entity init"""')
      initlines.extend(initbody)
      initlines.append('\n  self.Validate()')
      initcode = compile(''.join(initlines), '<ddd.shared>', 'exec')
      initfunc = FunctionType(
        InitCodeType(initcode.co_consts[ORD]), globals(), '__init__', None,
        cls.__init__.__closure__)
      super().__init__(cname, bases, namespace)
      cls.__init__ = initfunc
      cls.__slots__ = tuple(slots)
      if idname:
        annotations = cls.identity.__annotations__
        cls.identity = FunctionType(
          compile(
            f'def identity(self) -> ID: return self.{idname}\n',
            '<ddd.shared>',
            'single').co_consts[ORD],
          globals())
        cls.identity.__annotations__ = annotations
      else:
        cls.identity = MetaEntity.identity(cls)

  @staticmethod
  def identity(cls) -> Callable[[], ID]:
    def wrapper(_) -> NoReturn: raise NoIdentityError(cls)
    return wrapper

  def __repr__(cls):
    items = cls.__annotations__.items()
    props = ('{}={!r}'.format(name, prop) for name, prop in items)
    return '{}<{}>'.format(cls.__name__, ', '.join(props))


def InitCodeType(c):
  return CodeType(c.co_argcount, c.co_posonlyargcount,
                  c.co_kwonlyargcount, c.co_nlocals,
                  c.co_stacksize, c.co_flags, c.co_code, c.co_consts, c.co_names,
                  c.co_varnames, c.co_filename, c.co_name, c.co_firstlineno,
                  c.co_lnotab, c.co_freevars + ('__class__',), c.co_cellvars)


class Entity(Generic[T, ID], metaclass=MetaEntity):
  """TODO: Domain model entity."""

  def __init__(self):
    super().__init__()

  def Validate(self):
    """Execute domain member validations."""

  def identity(self) -> ID:
    return NotImplemented

  def SameIdentityAs(other: T) -> bool:
    return NotImplemented

  def __repr__(self):
    vals = ('{}={!r}'.format(m, getattr(self, m))
            for m in self.__annotations__)
    return '{}<{}>'.format(self.__class__.__name__, ', '.join(vals))

  @classmethod
  def Make(cls, **kwargs) -> Entity:
    entity = cls.__new__(cls)
    for key, value in kwargs.items():
      setattr(entity, key, value)
    return entity


class EntitySupport:
  def __eq__(self, other: Entity):
    return all(getattr(self, name) == getattr(other, name)
               for name in self.__slots__)

  def __hash__(self):
    return hash(self.identity())


class MetaValueObject(ABCMeta):
  def __init__(cls, cname, bases, namespace):
    if '__annotations__' in namespace:
      initlines = ['def __init__(self']
      initbody = []
      slots = []
      idname = None
      for name, kind in get_type_hints(cls).items():
        if isinstance(kind, _GenericAlias):
          kind = cast(_GenericAlias, kind).__origin__
        if isinstance(kind, _SpecialForm):
          initlines.append(f', {name}: {kind}')
        else:
          initlines.append(f', {name}: {kind.__name__}')
        initbody.append(f'\n  self.{name} = {name}')
        slots.append(name)
        if isinstance(kind, IdentityAlias):
          raise TypeError('Use Identity in ValueObject es invalid')
      initlines.append('):')
      initlines.extend(initbody)
      initlines.append('\n  self.Validate()')
      initcode = compile(''.join(initlines), '<ddd.shared>', 'exec')
      initfunc = FunctionType(
        InitCodeType(initcode.co_consts[ORD]), globals(), '__init__', None,
        cls.__init__.__closure__)
      super().__init__(cname, bases, namespace)
      cls.__init__ = initfunc
      cls.__slots__ = tuple(slots)

  def __repr__(cls):
    items = cls.__annotations__.items()
    props = ('{}={!r}'.format(name, prop) for name, prop in items)
    return '{}<{}>'.format(cls.__name__, ', '.join(props))


class ValueObject(metaclass=MetaValueObject):
  """TODO: Domain model value object."""

  def __init__(self):
    super().__init__()

  def Validate(self):
    """Execute domain member validations."""

  def __eq__(self, other: ValueObject) -> bool:
    return all(getattr(self, name) == getattr(other, name)
               for name in self.__slots__)

  def __repr__(self):
    vals = ('{}={!r}'.format(m, getattr(self, m))
            for m in self.__annotations__)
    return '{}<{}>'.format(self.__class__.__name__, ', '.join(vals))

  @classmethod
  def Make(cls, **kwargs) -> ValueObject:
    valueObject = cls.__new__(cls)
    for key, value in kwargs.items():
      setattr(valueObject, key, value)
    return valueObject


def _idcache(fn=None, /, *, typed=False):
  def wrapper(fn):
    cached = lru_cache(typed=typed)(fn)
    _idcache.clears.append(cached.cache_clear)

    @wraps(fn)
    def inner(*args, **kwds):
      try:
        return cached(*args, **kwds)
      except TypeError:
        pass
      return fn(*args, **kwds)
    return inner
  if fn is not None:
    return wrapper(fn)
  return wrapper


_idcache.clears = []


class Final:
  __slots__ = ('__weakref__',)

  def __init_subclass__(self, /, *_, **ks):
    if '_root' not in ks:
      raise TypeError("Cannot subclass")


class Immutable:
  __slots__ = ()
  def __copy__(self): return self
  def __deepcopy__(self, _): return self


class Identity(Immutable):
  """TODO: Domain model annotation for identity."""
  @_idcache
  def __class_getitem__(cls, member: Type[object]):
    return IdentityAlias(member)


class IdentityAlias(Final, Immutable, _root=True):
  __slots__ = ('_kind', '__name__')

  def __init__(self, kind: Type[object]):
    self._kind = kind
    self.__name__ = kind.__name__

  def __instancecheck__(self, obj): return self.__subclasscheck__(type(obj))

  def __subclasscheck__(self, cls): return cls == self._kind

  def __mro_entries__(self, bases): raise TypeError(
    f'Cannot subclass {self!r}')

  def __repr__(self): return f'{self._kind}'

  def __reduce__(self):
    return self._name

  def __call__(self, *args, **kwds):
    raise TypeError(f'Cannot instantiate {self!r}')

  def __instancecheck__(self, obj): raise TypeError('Cannot instance check')

  def __subclasscheck__(self, cls): raise TypeError('Cannot subclass check')


class Service(Stuff):
  """TODO: Domain service declaration."""


class Repository(Stuff):
  """TODO: Domain model entity repository."""


class Action:
  """TODO: Domain model action service."""

  @abstractmethod
  def Apply(self):
    raise NotImplementedError


class QuerySet:
  """TODO: Domain model query set application."""


class Error(Exception):
  """DDD Errors"""


class DomainError(Error):
  """Entities, repositories and services base error for user cases.
  This exceptions does not break the workflow."""


class NoIdentityError(DomainError):
  """Notify the use"""

  def __init__(self, cls):
    self._cls = cls

  def __str__(self): return f'{self._cls.__qualname__} without identity'


class RepositoryError(DomainError):
  """Find, Store, Delete, Remove errors"""


class ServiceError(DomainError):
  pass


class NotFoundError(RepositoryError):
  """Raised when Repository.Find doest not foun.
  Connection and transaction errors not included.
  """



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pathlib import Path

from django.conf import settings
from django.core.management.base import (BaseCommand, CommandError,
                                         CommandParser)

try:
  import autopep8
except ImportError as error:
  raise CommandError('autopep8 package not installed') from error

try:
  import isort
except ImportError as error:
  raise CommandError('isort package not installed') from error


class Command(BaseCommand):
  help = 'Apply pep8 format each python file.'

  def add_arguments(self, parser: CommandParser):
    parser.add_argument(
      '--file',
      '-f',
      help='Use when you want to apply format this file',
    )

  def handle(self, *args, **options):
    basedir = Path(settings.BASE_DIR)

    if not basedir.is_dir():
      raise CommandError(f'Invalid project directory: {basedir}')

    autopep8.DEFAULT_INDENT_SIZE = 2

    paths = (
      [options['file']] if options['file'] else basedir.glob('**/*.py')
    )

    for path in paths:
      name = str(path)
      autopep8.main(('autopep8', '-i', '-a', '-a', name))
      isort.file(name, quiet=True)



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import subprocess
from pathlib import Path

from django.conf import settings
from django.core.management.base import (BaseCommand, CommandError,
                                         CommandParser)

try:
  import coverage
except ImportError as error:
  raise CommandError('coverage package not installed') from error


class Command(BaseCommand):
  help = 'Run coverage tool.'

  def add_arguments(self, parser: CommandParser):
    parser.add_argument('test-module', help='Test module base')

  def handle(self, *args, **options):
    basedir = Path(settings.BASE_DIR)

    if not basedir.is_dir():
      raise CommandError(f'Invalid project directory: {basedir}')

    if not (basedir / 'manage.py').exists():
      raise CommandError(f'No project directory: {basedir}')

    testmodule = options['test-module']
    try:
      subprocess.run(
        ' '.join(
          (
            'coverage',
            'run',
            '--source=.',
            'manage.py',
            'test',
            testmodule,
          )
        ),
        shell=True,
        check=True,
      )
    except subprocess.CalledProcessError:
      return
    subprocess.run(['coverage', 'html'])



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Stuff definitions for enitty and stuffs designing."""

from __future__ import annotations

import builtins
from abc import ABC, ABCMeta
from sys import modules
from types import FunctionType
from typing import (Any, Callable, Dict, Generic, List, NamedTuple, Tuple,
                    Type, TypeVar, get_type_hints)

from .exceptions import BadMemberError

__all__ = ('Field',)

T = TypeVar('T')


class KindnameError(BadMemberError):
  """Raised when a member definition for a kindname can not be found. May also
  be raised when the kind is not a string."""


class _MetaStuff(ABCMeta):
  """Metaclass for Stuff

  To state members with the names needed. For example,

  .. code-block: python
    class Publication(Stuff):
      count: int

  ``Publication.count`` needs to know the name ``count`` defined.
  """

  def __init__(cls, name: str, bases: Tuple[str], namespace: Dict[str, Any]):
    super(_MetaStuff, cls).__init__(name, bases, namespace)
    cls._StateMembers()

  def __repr__(cls):
    fields = [f'{f.name}={f!r}' for _, f in sorted(cls._fields.items())]
    return f'{cls.__name__}<{", ".join(fields)}>'


class Stuff(ABC, metaclass=_MetaStuff):
  """A class describing domain stuffs.

  All stuff classes inheriting from :class:`Stuff` automatically have
  :class:`_MetaStuff` as their metaclass, so that members are stated
  properly after class definition."""

  __slots__ = ()

  _fields = {}
  _kinds = {}

  def __repr__(self):
    fields = [f'{f.name}={f!r}' for _, f in sorted(self._fields.items())]
    return f'{self._kindname()}<{", ".join(fields)}>'

  def ReflectionEquals(self, other: Stuff) -> bool:
    """Compare two stuff by fields."""
    if len(self._fields) != len(other._fields):
      return False
    return not any(name not in other._fields
                   or field != other._fields[name]
                   or getattr(self, name) != getattr(other, name)
                   for name, field in self._fields.items())

  @classmethod
  def _kindname(cls):
    """Return the kind name for this class.

    Defaults to ``cls.__name__``; users may override this to return a
    different name when stored than the name of the class.
    """
    return cls.__name__

  @classmethod
  def _StateMembers(cls):
    """State member calling their ``_FixState()`` method.

    .. note::

      Called by :class:`_MetaStuff`, but may also be called
      manually after dynamically updating a stuff class.

    Raises:
      KindnameError: When returned kind name from ``_kindname()`` is not a
        :class:`str`.
    """
    kindname = cls._kindname()
    if not isinstance(kindname, str):
      raise KindnameError(
        f'Class {cls.__name__} defines a ``_kindname()`` method that returns a'
        f' non-string ({kindname!r})')

    typeHints = get_type_hints(cls)
    if typeHints:
      cls._fields = {}
      cls._kinds = {}

      for name, field in typeHints.items():
        # TODO: back compatibility - remove after migration
        if not isinstance(field, Field):
          field = Field[field]
        field._FixState(cls, name)  # pylint:disable=protected-access
        cls._fields[name] = field
      cls._UpdateKinds()

      _StateNewAttribute(
        cls, '__init__',
        _InitFn(
          cls._fields.values(),
          'self', modules[cls.__module__].__dict__
          if cls.__module__ in modules else{},))

  @classmethod
  def _UpdateKinds(cls: Type[Stuff]):
    """Update kinds dict to include this class."""
    cls._kinds[cls._kindname()] = cls


def _InitFn(fields: List[Field], selfname: str,
            globalns: Dict[str, Type[object]]):
  """Init Function."""
  localns = {f'_type_{f.name}': f.kind for f in fields}
  lines = [f'{selfname}.{f.name}={f.name}' for f in fields]
  params = [f'{f.name}:{f.kind.__name__}' for f in fields]

  return _NewFn('__init__',
                [selfname] + params,
                lines,
                localns=localns,
                globalns=globalns,
                return_type=bool)


def _StateNewAttribute(
    cls: Type[object],
    name: str, value: Callable[..., None]):
  """Never overwrites an existing member."""
  if name in cls.__dict__:
    raise ValueError(f'{cls}: has name={name} with value={value}')
  _StateQualname(cls, value)
  setattr(cls, name, value)


def _StateQualname(cls, value):
  """_NewFn uses proper __qualname__."""
  if isinstance(value, FunctionType):
    value.__qualname__ = f"{cls.__qualname__}.{value.__name__}"
  return value


def _NewFn(
    name: str, args: List[str],
    body: List[str],
    *, globalns: Dict[str, Type[object]] = None,
    localns: Dict[str, Type[object]] = None, return_type: Type[object] = None):
  """Create new function. localns mutates on exec()."""
  if localns is None:
    localns = {}
  if 'BUILTINS' not in localns:
    localns['BUILTINS'] = builtins
  return_annotation = ''
  if return_type:
    localns['_return_type'] = return_type
    return_annotation = '->_return_type'
  args = ','.join(args)
  body = '\n'.join(f'  {b}' for b in body)
  txt = f' def {name}({args}){return_annotation}:\n{body}'
  localvars = ', '.join(localns.keys())
  txt = f"def __new_fn__({localvars}):\n{txt}\n return {name}"
  ns = {}
  exec(txt, globalns, ns)  # pylint:disable=exec-used
  return ns['__new_fn__'](**localns)


class Field(Generic[T]):
  """A class describing a persisted kind member."""

  __slots__ = ('kind', 'validator', 'pkind', 'name')

  def __init__(self, kind: Type[T], validator: Callable[[T], None] = None):
    self.kind = kind
    self.validator = validator
    self.name = None
    self.pkind = None

  def __repr__(self):
    return f'Field[{self.kind} name={self.name},' \
           f' pkind={self.pkind._kindname()}]'

  def __call__(self, *args, **kwds):
    raise TypeError(f"Cannot instantiate {self!r}")

  def __eq__(self, other: Field) -> bool:
    return (other and isinstance(other, Field)
            and self.kind is other.kind and self.name == other.name)

  def _FixState(self, pkind: Type[T], name: str):
    """Helper called to tell member its name."""
    self.pkind = pkind
    self.name = name

  def __class_getitem__(
      cls, item: T | Tuple[Type[T], Callable[[T], None] | Arg]):
    return cls(*item) if isinstance(item, Arg) else cls(item)


class Arg(NamedTuple):
  """To describe Field definition."""
  kind: Type[T]
  validator: Callable[[T], None] = None



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""This module is under construction.
It's used to mark the domain role for classes and models defined in the domain.
"""

from __future__ import annotations

from .stuff import Field

__all__ = ('Identity',)


class Identity(Field):  # pylint:disable=too-few-public-methods
  """Every class that inherits from EntitySupport must have exactly a identity.
  """



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from django.template import Library, Node
from django.template.base import FilterExpression, Parser, Token
from django.template.loader_tags import construct_relative_path

register = Library()


class NeomImportNode(Node):
  def __init__(self, template: FilterExpression, *args, **kwargs):
    self.template = template
    super().__init__(*args, **kwargs)

  def render(self, context):
    subpath = self.template.resolve(context)
    fullpath = (
      construct_relative_path(self.origin.template_name, subpath),
    )

    cache = context.render_context.dicts[0].setdefault(self, {})
    template = cache.get(fullpath)

    if not template:
      template = context.template.engine.select_template(fullpath)
      cache[fullpath] = template

    return template.render(context)


@register.tag
def neom_import(parser: Parser, token: Token):
  bits = token.split_contents()
  if len(bits) < 2:
    raise template.TemplateSyntaxError(
      f'{bits[0]} tag takes at least one argument: the asset path'
    )
  bits[1] = construct_relative_path(parser.origin.template_name, bits[1])
  return NeomImportNode(parser.compile_filter(bits[1]))



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Staff classes.
Common used in entities or value objects."""

from __future__ import annotations

import dataclasses
import re
from typing import Final, Generic, Optional, TypeVar, Type

from .shared import ValueObject, _idcache


class Phone(ValueObject):
  """Compose phone."""
  country: Optional[int]
  area: Optional[int]
  number: Final[int]


class Mobile(Phone):
  """Mobile number as integer."""
  REGEX = r'^\(\+(\d{2})\) (\d{3}(?:-\d{3}){2})$'

  def __str__(self):
    part1 = int(self.number / 10 ** 6)
    part2 = int((self.number % 10 ** 6) / 10 ** 3)
    part3 = int(self.number % 10 ** 3)
    return f'(+{self.country}) {part1}-{part2}-{part3}'

  @staticmethod
  def Make(fmt: str) -> Mobile:
    """Make from string format number."""
    match = re.match(Mobile.REGEX, fmt)
    if not match:
      raise ValueError(f'Invalid mobile phone {fmt}. User (+xx) xxx-xxx-xxx')

    groups = match.groups()
    country = int(groups[0])
    area = None
    number = int(groups[1].replace('-', ''))

    return Mobile(country=country, area=area, number=number)


class Email(ValueObject):
  """Email parsed."""
  address: str

  REGEX = r'^\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b$'

  def Validate(self):
    """Validate address regex."""
    if not re.match(self.REGEX, self.address):
      raise ValueError(f'Invalid email address format {self.address}')

  def __str__(self):
    return self.address


K = TypeVar('K')

class Key(ValueObject, Generic[K]):
  """Entity key."""

  k: K


  def __hash__(self) -> int:
    return hash(self.k)

  def __repr__(self):
    return f'Key<{self.k}: {K}={K}>'

  def _keyType(self) -> Type[K]:
    """key type."""
    return self.__orig_class__[0]

  @classmethod
  def Next(cls):
    """Generate next valid key."""
    print(cls(0))

  @_idcache
  def __class_getitem__(self, k: K):
    return Key

IntKey = Key[int]
